@@ -296,6 +296,7 @@ static struct show_image_info show_infos[] = {
 	SHOW_VERT(FS),
 	SHOW_VERT(GHOST_FILE),
 	SHOW_VERT(MM),
+	SHOW_VERT(CGROUP),
 
 	SHOW_PLAINS(REG_FILE),
 	SHOW_PLAINS(NS_FILE),
@@ -303,6 +304,7 @@ static struct show_image_info show_infos[] = {
 	SHOW_PLAIN(EVENTPOLL_FILE),
 	SHOW_PLAIN(EVENTPOLL_TFD),
 	SHOW_PLAIN(SIGNALFD),
+	SHOW_PLAIN(TIMERFD),
 	SHOW_PLAIN(INOTIFY_FILE),
 	SHOW_PLAIN(INOTIFY_WD),
 	SHOW_PLAIN(FANOTIFY_FILE),
@@ -316,12 +318,12 @@ static struct show_image_info show_infos[] = {
 	SHOW_PLAINS(MNT),
 	SHOW_PLAINS(TTY_FILE),
 	SHOW_PLAIN(TTY_INFO),
-	SHOW_PLAINS(FILE_LOCK),
 	SHOW_PLAIN(RLIMIT),
 	SHOW_PLAIN(TUNFILE),
 	SHOW_PLAINS(EXT_FILE),
 	SHOW_PLAIN(IRMAP_CACHE),
 
+	{ FILE_LOCKS_MAGIC,	PB_FILE_LOCK,		false,	NULL, "3:%u", },
 	{ TCP_STREAM_MAGIC,	PB_TCP_STREAM,		true,	show_tcp_stream, "1:%u 2:%u 3:%u 4:%u 12:%u", },
 	{ STATS_MAGIC,		PB_STATS,		true,	NULL, "1.1:%u 1.2:%u 1.3:%u 1.4:%u 1.5:%Lu 1.6:%Lu 1.7:%Lu 1.8:%u", },
 	{ FDINFO_MAGIC,		PB_FDINFO,		false,	NULL, "flags:%#o fd:%d", },
@@ -443,6 +445,16 @@ static int cr_show_pstree_item(struct pstree_item *item)
 			cr_parse_fd(fdset_fd(cr_fdset, i), fdset_template[i].magic);
 		}
 
+	i = open_image(CR_FD_RLIMIT, O_SHOW | O_OPT, item->pid.virt);
+	if (i >= 0) {
+		pr_msg("* ");
+		pr_msg(fdset_template[CR_FD_RLIMIT].fmt, item->pid.virt);
+		pr_msg(":\n");
+
+		cr_parse_fd(i, RLIMIT_MAGIC);
+		close(i);
+	}
+
 	if (pb_read_one(fdset_fd(cr_fdset, CR_FD_IDS), &ids, PB_IDS) > 0) {
 		i = open_image(CR_FD_FDINFO, O_SHOW, ids->files_id);
 		if (i >= 0) {
