@@ -34,6 +34,8 @@
 #include "file-lock.h"
 #include "cr-service.h"
 #include "plugin.h"
+#include "mount.h"
+#include "cgroup.h"
 
 struct cr_options opts;
 
@@ -45,6 +47,11 @@ void init_opts(void)
 	opts.final_state = TASK_DEAD;
 	INIT_LIST_HEAD(&opts.veth_pairs);
 	INIT_LIST_HEAD(&opts.scripts);
+	INIT_LIST_HEAD(&opts.ext_mounts);
+	INIT_LIST_HEAD(&opts.new_cgroup_roots);
+
+	opts.cpu_cap = CPU_CAP_ALL;
+	opts.manage_cgroups = false;
 }
 
 static int parse_ns_string(const char *ptr)
@@ -75,16 +82,54 @@ bad_ns:
 	return -1;
 }
 
+static int parse_cpu_cap(struct cr_options *opts, const char *optarg)
+{
+	bool inverse = false;
+
+#define ____cpu_set_cap(__opts, __cap, __inverse)	\
+	do {						\
+		if ((__inverse))			\
+			(__opts)->cpu_cap &= ~(__cap);	\
+		else					\
+			(__opts)->cpu_cap |=  (__cap);	\
+	} while (0)
+
+	for (; *optarg; optarg++) {
+		if (optarg[0] == '^') {
+			inverse = !inverse;
+			continue;
+		} else if (optarg[0] == ',') {
+			inverse = false;
+			continue;
+		}
+
+		if (!strncmp(optarg, "fpu", 3))
+			____cpu_set_cap(opts, CPU_CAP_FPU, inverse);
+		if (!strncmp(optarg, "all", 3))
+			____cpu_set_cap(opts, CPU_CAP_ALL, inverse);
+		else
+			goto Esyntax;
+	}
+#undef ____cpu_set_cap
+
+	return 0;
+
+Esyntax:
+	pr_err("Unknown FPU mode `%s' selected\n", optarg);
+	return -1;
+}
+
 int main(int argc, char *argv[])
 {
 	pid_t pid = 0, tree_id = 0;
 	int ret = -1;
 	bool usage_error = true;
+	bool has_exec_cmd = false;
 	int opt, idx;
 	int log_level = LOG_UNSET;
 	char *imgs_dir = ".";
 	char *work_dir = NULL;
-	static const char short_opts[] = "dsRf:F:t:p:hcD:o:n:v::xVr:jlW:L:";
+	static const char short_opts[] = "dsRf:F:t:p:hcD:o:n:v::xVr:jlW:L:M:";
 	static struct option long_opts[] = {
 		{ "tree", required_argument, 0, 't' },
 		{ "pid", required_argument, 0, 'p' },
@@ -102,32 +147,39 @@ int main(int argc, char *argv[])
 		{ "root", required_argument, 0, 'r' },
 		{ USK_EXT_PARAM, no_argument, 0, 'x' },
 		{ "help", no_argument, 0, 'h' },
-		{ SK_EST_PARAM, no_argument, 0, 42 },
-		{ "close", required_argument, 0, 43 },
-		{ "log-pid", no_argument, 0, 44},
+		{ SK_EST_PARAM, no_argument, 0, 1042 },
+		{ "close", required_argument, 0, 1043 },
+		{ "log-pid", no_argument, 0, 1044},
 		{ "version", no_argument, 0, 'V'},
-		{ "evasive-devices", no_argument, 0, 45},
-		{ "pidfile", required_argument, 0, 46},
-		{ "veth-pair", required_argument, 0, 47},
-		{ "action-script", required_argument, 0, 49},
-		{ LREMAP_PARAM, no_argument, 0, 41},
+		{ "evasive-devices", no_argument, 0, 1045},
+		{ "pidfile", required_argument, 0, 1046},
+		{ "veth-pair", required_argument, 0, 1047},
+		{ "action-script", required_argument, 0, 1049},
+		{ LREMAP_PARAM, no_argument, 0, 1041},
 		{ OPT_SHELL_JOB, no_argument, 0, 'j'},
 		{ OPT_FILE_LOCKS, no_argument, 0, 'l'},
-		{ "page-server", no_argument, 0, 50},
-		{ "address", required_argument, 0, 51},
-		{ "port", required_argument, 0, 52},
-		{ "prev-images-dir", required_argument, 0, 53},
-		{ "ms", no_argument, 0, 54},
-		{ "track-mem", no_argument, 0, 55},
-		{ "auto-dedup", no_argument, 0, 56},
+		{ "page-server", no_argument, 0, 1050},
+		{ "address", required_argument, 0, 1051},
+		{ "port", required_argument, 0, 1052},
+		{ "prev-images-dir", required_argument, 0, 1053},
+		{ "ms", no_argument, 0, 1054},
+		{ "track-mem", no_argument, 0, 1055},
+		{ "auto-dedup", no_argument, 0, 1056},
 		{ "libdir", required_argument, 0, 'L'},
+		{ "cpu-cap", required_argument, 0, 1057},
+		{ "force-irmap", no_argument, 0, 1058},
+		{ "ext-mount-map", required_argument, 0, 'M'},
+		{ "exec-cmd", no_argument, 0, 1059},
+		{ "manage-cgroups", no_argument, 0, 1060},
+		{ "cgroup-root", required_argument, 0, 1061},
 		{ },
 	};
 
 	BUILD_BUG_ON(PAGE_SIZE != PAGE_IMAGE_SIZE);
 
 	cr_pb_init();
-	restrict_uid(getuid(), getgid());
+	if (restrict_uid(getuid(), getgid()))
+		return 1;
 
 	if (argc < 2)
 		goto usage;
@@ -137,6 +189,17 @@ int main(int argc, char *argv[])
 	if (init_service_fd())
 		return 1;
 
+	if (!strcmp(argv[1], "swrk")) {
+		/*
+		 * This is to start criu service worker from libcriu calls.
+		 * The usage is "criu swrk <fd>" and is not for CLI/scripts.
+		 * The arguments semantics can change at any tyme with the
+		 * corresponding lib call change.
+		 */
+		opts.swrk_restore = true;
+		return cr_service_work(atoi(argv[2]));
+	}
+
 	while (1) {
 		idx = -1;
 		opt = getopt_long(argc, argv, short_opts, long_opts, &idx);
@@ -203,15 +266,15 @@ int main(int argc, char *argv[])
 			} else
 				log_level++;
 			break;
-		case 41:
+		case 1041:
 			pr_info("Will allow link remaps on FS\n");
 			opts.link_remap_ok = true;
 			break;
-		case 42:
+		case 1042:
 			pr_info("Will dump TCP connections\n");
 			opts.tcp_established_ok = true;
 			break;
-		case 43: {
+		case 1043: {
 			int fd;
 
 			fd = atoi(optarg);
@@ -219,16 +282,16 @@ int main(int argc, char *argv[])
 			close(fd);
 			break;
 		}
-		case 44:
+		case 1044:
 			opts.log_file_per_pid = 1;
 			break;
-		case 45:
+		case 1045:
 			opts.evasive_devices = true;
 			break;
-		case 46:
+		case 1046:
 			opts.pidfile = optarg;
 			break;
-		case 47:
+		case 1047:
 			{
 				char *aux;
 
@@ -241,7 +304,7 @@ int main(int argc, char *argv[])
 					return 1;
 			}
 			break;
-		case 49:
+		case 1049:
 			{
 				struct script *script;
 
@@ -253,13 +316,13 @@ int main(int argc, char *argv[])
 				list_add(&script->node, &opts.scripts);
 			}
 			break;
-		case 50:
+		case 1050:
 			opts.use_page_server = true;
 			break;
-		case 51:
+		case 1051:
 			opts.addr = optarg;
 			break;
-		case 52:
+		case 1052:
 			opts.ps_port = htons(atoi(optarg));
 			if (!opts.ps_port)
 				goto bad_arg;
@@ -270,21 +333,65 @@ int main(int argc, char *argv[])
 		case 'l':
 			opts.handle_file_locks = true;
 			break;
-		case 53:
+		case 1053:
 			opts.img_parent = optarg;
 			break;
-		case 55:
+		case 1055:
 			opts.track_mem = true;
 			break;
-		case 56:
+		case 1056:
 			opts.auto_dedup = true;
 			break;
-		case 54:
+		case 1057:
+			if (parse_cpu_cap(&opts, optarg))
+				goto usage;
+			break;
+		case 1058:
+			opts.force_irmap = true;
+			break;
+		case 1054:
 			opts.check_ms_kernel = true;
 			break;
 		case 'L':
 			opts.libdir = optarg;
 			break;
+		case 1059:
+			has_exec_cmd = true;
+			break;
+		case 1060:
+			opts.manage_cgroups = true;
+			break;
+		case 1061:
+			{
+				char *path, *ctl;
+
+				path = strchr(optarg, ':');
+				if (path) {
+					*path = '\0';
+					path++;
+					ctl = optarg;
+				} else {
+					path = optarg;
+					ctl = NULL;
+				}
+
+				if (new_cg_root_add(ctl, path))
+					return -1;
+			}
+			break;
+		case 'M':
+			{
+				char *aux;
+
+				aux = strchr(optarg, ':');
+				if (aux == NULL)
+					goto bad_arg;
+
+				*aux = '\0';
+				if (ext_mount_add(optarg, aux + 1))
+					return 1;
+			}
+			break;
 		case 'V':
 			pr_msg("Version: %s\n", CRIU_VERSION);
 			if (strcmp(CRIU_GITID, "0"))
@@ -306,6 +413,27 @@ int main(int argc, char *argv[])
 		goto usage;
 	}
 
+	if (has_exec_cmd) {
+		if (argc - optind <= 1) {
+			pr_msg("Error: --exec-cmd requires a command\n");
+			goto usage;
+		}
+
+		if (strcmp(argv[optind], "restore")) {
+			pr_msg("Error: --exec-cmd is available for the restore command only\n");
+			goto usage;
+		}
+
+		if (opts.restore_detach) {
+			pr_msg("Error: --restore-detached and --exec-cmd cannot be used together\n");
+			goto usage;
+		}
+
+		opts.exec_cmd = xmalloc((argc - optind) * sizeof(char *));
+		memcpy(opts.exec_cmd, &argv[optind + 1], (argc - optind - 1) * sizeof(char *));
+		opts.exec_cmd[argc - optind - 1] = NULL;
+	}
+
 	/* We must not open imgs dir, if service is called */
 	if (strcmp(argv[optind], "service")) {
 		ret = open_image_dir(imgs_dir);
@@ -342,7 +470,16 @@ int main(int argc, char *argv[])
 	if (!strcmp(argv[optind], "restore")) {
 		if (tree_id)
 			pr_warn("Using -t with criu restore is obsoleted\n");
-		return cr_restore_tasks() != 0;
+
+		ret = cr_restore_tasks();
+		if (ret == 0 && opts.exec_cmd) {
+			close_pid_proc();
+			execvp(opts.exec_cmd[0], opts.exec_cmd);
+			pr_perror("Failed to exec command %s", opts.exec_cmd[0]);
+			ret = 1;
+		}
+
+		return ret != 0;
 	}
 
 	if (!strcmp(argv[optind], "show"))
@@ -410,6 +547,10 @@ usage:
 "     --pidfile FILE     write root task, service or page-server pid to FILE\n"
 "  -W|--work-dir DIR     directory to cd and write logs/pidfiles/stats to\n"
 "                        (if not specified, value of --images-dir is used)\n"
+"     --cpu-cap CAP      require certain cpu capability. CAP: may be one of:\n"
+"                        'fpu','all'. To disable capability, prefix it with '^'.\n"
+"     --exec-cmd         execute the command specified after '--' on successful\n"
+"                        restore making it the parent of the restored process\n"
 "\n"
 "* Special resources support:\n"
 "  -x|--" USK_EXT_PARAM "      allow external unix connections\n"
@@ -423,6 +564,14 @@ usage:
 "  -j|--" OPT_SHELL_JOB "        allow to dump and restore shell jobs\n"
 "  -l|--" OPT_FILE_LOCKS "       handle file locks, for safety, only used for container\n"
 "  -L|--libdir           path to a plugin directory (by default " CR_PLUGIN_DEFAULT ")\n"
+"  --force-irmap         force resolving names for inotify/fsnotify watches\n"
+"  -M|--ext-mount-map KEY:VALUE\n"
+"                        add external mount mapping\n"
+"  --manage-cgroups      dump or restore cgroups the process is in\n"
+"  --cgroup-root [controller:]/newroot\n"
+"                        change the root cgroup the controller will be\n"
+"                        installed into. No controller means that root is the\n"
+"                        default for all controllers not specified.\n"
 "\n"
 "* Logging:\n"
 "  -o|--log-file FILE    log file name\n"
@@ -439,6 +588,8 @@ usage:
 "  --page-server         send pages to page server (see options below as well)\n"
 "  --auto-dedup          when used on dump it will deduplicate \"old\" data in\n"
 "                        pages images of previous dump\n"
+"                        when used on restore, as soon as page is restored, it\n"
+"                        will be punched from the image.\n"
 "\n"
 "Page/Service server options:\n"
 "  --address ADDR        address of server or service\n"
