@@ -13,6 +13,7 @@
 #include <sys/types.h>
 #include <sys/inotify.h>
 #include <sys/vfs.h>
+#include <linux/magic.h>
 #include <sys/wait.h>
 #include <sys/poll.h>
 #include <sys/mman.h>
@@ -37,9 +38,13 @@
 #include "list.h"
 #include "lock.h"
 #include "irmap.h"
+#include "cr_options.h"
+#include "namespaces.h"
+#include "pstree.h"
 
 #include "protobuf.h"
 #include "protobuf/fsnotify.pb-c.h"
+#include "protobuf/mnt.pb-c.h"
 
 #undef	LOG_PREFIX
 #define LOG_PREFIX "fsnotify: "
@@ -110,7 +115,7 @@ static int open_handle(unsigned int s_dev, unsigned long i_ino,
 
 	mntfd = open_mount(s_dev);
 	if (mntfd < 0) {
-		pr_perror("Mount root for 0x%08x not found\n", s_dev);
+		pr_err("Mount root for 0x%08x not found\n", s_dev);
 		goto out;
 	}
 
@@ -129,14 +134,50 @@ out:
 int check_open_handle(unsigned int s_dev, unsigned long i_ino,
 		FhEntry *f_handle)
 {
-	int fd;
+	int fd = -1;
 	char *path;
 
 	fd = open_handle(s_dev, i_ino, f_handle);
 	if (fd >= 0) {
-		close(fd);
+		struct mount_info *mi;
+
 		pr_debug("\tHandle %x:%lx is openable\n", s_dev, i_ino);
-		return 0;
+
+		mi = lookup_mnt_sdev(s_dev);
+		if (mi == NULL) {
+			pr_err("Unable to lookup a mount by dev %x\n", s_dev);
+			goto err;
+		}
+
+		/*
+		 * Inode numbers are not restored for tmpfs content, but we can
+		 * get file names, becasue tmpfs cache is not pruned.
+		 */
+		if ((mi->fstype->code == FSTYPE__TMPFS) ||
+				(mi->fstype->code == FSTYPE__DEVTMPFS)) {
+			char p[PATH_MAX];
+
+			if (read_fd_link(fd, p, sizeof(p)) < 0)
+				goto err;
+
+			path = xstrdup(p);
+			if (path == NULL)
+				goto err;
+
+			goto out;
+		}
+
+		if (!opts.force_irmap)
+			/*
+			 * If we're not forced to do irmap, then
+			 * say we have no path for watch. Otherwise
+			 * do irmap scan even if the handle is
+			 * working.
+			 *
+			 * FIXME -- no need to open-by-handle if
+			 * we are in force-irmap and not on tempfs
+			 */
+			goto out_nopath;
 	}
 
 	pr_warn("\tHandle %x:%lx cannot be opened\n", s_dev, i_ino);
@@ -145,10 +186,15 @@ int check_open_handle(unsigned int s_dev, unsigned long i_ino,
 		pr_err("\tCan't dump that handle\n");
 		return -1;
 	}
-
+out:
 	pr_debug("\tDumping %s as path for handle\n", path);
 	f_handle->path = path;
+out_nopath:
+	close_safe(&fd);
 	return 0;
+err:
+	close_safe(&fd);
+	return -1;
 }
 
 static int dump_inotify_entry(union fdinfo_entries *e, void *arg)
@@ -276,7 +322,8 @@ static int pre_dump_fanotify_entry(union fdinfo_entries *e, void *arg)
 
 static int pre_dump_one_fanotify(int pid, int lfd)
 {
-	return parse_fdinfo_pid(pid, lfd, FD_TYPES__FANOTIFY, pre_dump_fanotify_entry, NULL);
+	struct fsnotify_params fsn_params = { };
+	return parse_fdinfo_pid(pid, lfd, FD_TYPES__FANOTIFY, pre_dump_fanotify_entry, &fsn_params);
 }
 
 const struct fdtype_ops fanotify_dump_ops = {
@@ -292,17 +339,24 @@ static char *get_mark_path(const char *who, struct file_remap *remap,
 	char *path = NULL;
 
 	if (remap) {
+		int mntns_root;
+
+		mntns_root = mntns_get_root_by_mnt_id(remap->mnt_id);
+
 		pr_debug("\t\tRestore %s watch for 0x%08x:0x%016lx (via %s)\n",
 			 who, s_dev, i_ino, remap->path);
-		return remap->path;
-	}
+		*target = openat(mntns_root, remap->path, O_PATH);
+	} else if (f_handle->path) {
+		int  mntns_root;
+
+		/* irmap cache is collected in the root namespaces. */
+		mntns_root = mntns_get_root_by_mnt_id(-1);
 
-	if (f_handle->path) {
 		pr_debug("\t\tRestore with path hint %s\n", f_handle->path);
-		return f_handle->path;
-	}
+		*target = openat(mntns_root, f_handle->path, O_PATH);
+	} else
+		*target = open_handle(s_dev, i_ino, f_handle);
 
-	*target = open_handle(s_dev, i_ino, f_handle);
 	if (*target < 0)
 		goto err;
 
@@ -383,9 +437,9 @@ static int restore_one_fanotify(int fd, struct fsnotify_mark_info *mark)
 	if (fme->type == MARK_TYPE__MOUNT) {
 		struct mount_info *m;
 
-		m = lookup_mnt_sdev(fme->s_dev);
+		m = lookup_mnt_id(fme->me->mnt_id);
 		if (!m) {
-			pr_err("Can't find mount s_dev %x\n", fme->s_dev);
+			pr_err("Can't find mount mnt_id %x\n", fme->me->mnt_id);
 			return -1;
 		}
 
