@@ -13,7 +13,7 @@ extern int check_img_inventory(void);
 extern int write_img_inventory(void);
 extern void kill_inventory(void);
 
-#define LAST_PID_PATH		"/proc/sys/kernel/ns_last_pid"
+#define LAST_PID_PATH		"sys/kernel/ns_last_pid"
 
 extern int cr_dump_tasks(pid_t pid);
 extern int cr_pre_dump_tasks(pid_t pid);
@@ -24,7 +24,7 @@ extern int cr_check(void);
 extern int cr_exec(int pid, char **opts);
 extern int cr_dedup(void);
 
-extern void restrict_uid(unsigned int uid, unsigned int gid);
+extern int restrict_uid(unsigned int uid, unsigned int gid);
 struct proc_status_creds;
 extern bool may_dump(struct proc_status_creds *);
 struct _CredsEntry;
