@@ -11,17 +11,12 @@ enum {
 	 */
 
 	_CR_FD_TASK_FROM,
-	CR_FD_FILE_LOCKS,
 	CR_FD_CORE,
 	CR_FD_IDS,
 	CR_FD_MM,
 	CR_FD_SIGACT,
-	CR_FD_ITIMERS,
-	CR_FD_POSIX_TIMERS,
 	CR_FD_CREDS,
 	CR_FD_FS,
-	CR_FD_RLIMIT,
-	CR_FD_SIGNAL,
 	_CR_FD_TASK_TO,
 
 	CR_FD_PAGEMAP,
@@ -77,18 +72,28 @@ enum {
 	CR_FD_FANOTIFY_FILE,
 	CR_FD_FANOTIFY_MARK,
 	CR_FD_TUNFILE,
+	CR_FD_CGROUP,
+	CR_FD_TIMERFD,
+	CR_FD_FILE_LOCKS,
 	_CR_FD_GLOB_TO,
 
-	CR_FD_TMPFS,
+	CR_FD_TMPFS_IMG,
+	CR_FD_TMPFS_DEV,
 	CR_FD_PAGES,
-	CR_FD_PSIGNAL,
 
 	CR_FD_VMAS,
 	CR_FD_PAGES_OLD,
 	CR_FD_SHM_PAGES_OLD,
+	CR_FD_RLIMIT,
+	CR_FD_ITIMERS,
+	CR_FD_POSIX_TIMERS,
+	CR_FD_FILE_LOCKS_PID,
 
 	CR_FD_IRMAP_CACHE,
 
+	CR_FD_SIGNAL,
+	CR_FD_PSIGNAL,
+
 	CR_FD_MAX
 };
 
