@@ -1,6 +1,8 @@
 #ifndef __CR_KCMP_IDS_H__
 #define __CR_KCMP_IDS_H__
 
+#include "kcmp.h"
+
 struct kid_tree {
 	struct rb_root root;
 	unsigned kcmp_type;
@@ -21,8 +23,8 @@ struct kid_elem {
 	unsigned idx;
 };
 
-u32 kid_generate_gen(struct kid_tree *tree,
-		struct kid_elem *elem, int *new_id);
-void kid_show_tree(struct kid_tree *tree);
+extern u32 kid_generate_gen(struct kid_tree *tree,
+			    struct kid_elem *elem, int *new_id);
+extern void kid_show_tree(struct kid_tree *tree);
 
 #endif /* __CR_KCMP_IDS_H__ */
