@@ -50,6 +50,8 @@ enum {
 	PB_SIGINFO,
 	PB_TUNFILE,
 	PB_IRMAP_CACHE,
+	PB_CGROUP,
+	PB_TIMERFD,
 
 	/* PB_AUTOGEN_STOP */
 
@@ -57,9 +59,9 @@ enum {
 	PB_IDS,
 	PB_SIGACT,
 	PB_NETDEV,
-	PB_REMAP_FPATH,
+	PB_REMAP_FPATH,		/* 50 */
 	PB_SK_QUEUES,
-	PB_IPCNS_MSG,		/* 50 */
+	PB_IPCNS_MSG,
 	PB_IPCNS_MSG_ENT,
 
 	PB_MAX,
