@@ -130,17 +130,28 @@ extern int open_pid_proc(pid_t pid);
 extern int close_pid_proc(void);
 extern int set_proc_fd(int fd);
 
+/*
+ * Values for pid argument of the proc opening routines below.
+ * SELF would open file under /proc/self
+ * GEN would open a file under /proc itself
+ * NONE is internal, don't use it ;)
+ */
+
+#define PROC_SELF	0
+#define PROC_GEN	-1
+#define PROC_NONE	-2
+
 extern int do_open_proc(pid_t pid, int flags, const char *fmt, ...);
 
-#define __open_proc(pid, flags, fmt, ...)			\
-	({							\
-		int __fd = do_open_proc(pid, flags,		\
-					fmt, ##__VA_ARGS__);	\
-		if (__fd < 0)					\
-			pr_perror("Can't open /proc/%d/" fmt,	\
-					pid, ##__VA_ARGS__);	\
-								\
-		__fd;						\
+#define __open_proc(pid, flags, fmt, ...)				\
+	({								\
+		int __fd = do_open_proc(pid, flags,			\
+					fmt, ##__VA_ARGS__);		\
+		if (__fd < 0)						\
+			pr_perror("Can't open %d/" fmt " on procfs",	\
+					pid, ##__VA_ARGS__);		\
+									\
+		__fd;							\
 	})
 
 /* int open_proc(pid_t pid, const char *fmt, ...); */
@@ -161,7 +172,7 @@ extern int do_open_proc(pid_t pid, int flags, const char *fmt, ...);
 			__d = fdopendir(__fd);				\
 			if (__d == NULL)				\
 				pr_perror("Can't fdopendir %d "		\
-					"(/proc/%d/" fmt ")",		\
+					"(%d/" fmt " on procfs)",	\
 					__fd, pid, ##__VA_ARGS__);	\
 		}							\
 		__d;							\
@@ -177,7 +188,7 @@ extern int do_open_proc(pid_t pid, int flags, const char *fmt, ...);
 			__f = fdopen(__fd, "r");			\
 			if (__f == NULL)				\
 				pr_perror("Can't fdopen %d "		\
-					"(/proc/%d/" fmt ")",		\
+					"(%d/" fmt " on procfs)",	\
 					__fd, pid, ##__VA_ARGS__);	\
 		}							\
 		__f;							\
@@ -211,11 +222,8 @@ static inline dev_t kdev_to_odev(u32 kdev)
 	 */
 	unsigned major = kdev_major(kdev);
 	unsigned minor = kdev_minor(kdev);
-#if BITS_PER_LONG == 32
-	return (major << 8) | minor;
-#else
-	return (minor & 0xff) | (major << 8) | ((minor & ~0xff) << 12);
-#endif
+
+	return makedev(major, minor);
 }
 
 extern int copy_file(int fd_in, int fd_out, size_t bytes);
@@ -277,4 +285,35 @@ extern int read_fd_link(int lfd, char *buf, size_t size);
 
 int vaddr_to_pfn(unsigned long vaddr, u64 *pfn);
 
+/*
+ * Check whether @str starts with @sub
+ */
+static inline bool strstartswith(const char *str, const char *sub)
+{
+	while (1) {
+		if (*sub == '\0') /* end of sub -- match */
+			return true;
+		if (*str == '\0') /* end of str, sub is NOT ended -- miss */
+			return false;
+		if (*str != *sub)
+			return false;
+
+		str++;
+		sub++;
+	}
+}
+
+/*
+ * mkdir -p
+ */
+int mkdirp(const char *path);
+
+/*
+ * Tests whether a path is a prefix of another path. This is different than
+ * strstartswith because "/foo" is _not_ a path prefix of "/foobar", since they
+ * refer to different directories.
+ */
+bool is_path_prefix(const char *path, const char *prefix);
+FILE *fopenat(int dirfd, char *path, char *cflags);
+void split(char *str, char token, char ***out, int *n);
 #endif /* __CR_UTIL_H__ */
