@@ -7,6 +7,7 @@
 #include <errno.h>
 #include <sys/stat.h>
 #include <string.h>
+#include <ctype.h>
 #include <linux/fs.h>
 
 #include "asm/types.h"
@@ -24,8 +25,11 @@
 #include "vma.h"
 
 #include "proc_parse.h"
+#include "cr_options.h"
+#include "sysfs_parse.h"
 #include "protobuf.h"
 #include "protobuf/fdinfo.pb-c.h"
+#include "protobuf/mnt.pb-c.h"
 
 #include <stdlib.h>
 
@@ -92,6 +96,8 @@ static bool is_vma_range_fmt(char *line)
 static int parse_vmflags(char *buf, struct vma_area *vma_area)
 {
 	char *tok;
+	bool shared = false;
+	bool maywrite = false;
 
 	if (!buf[0])
 		return 0;
@@ -103,6 +109,12 @@ static int parse_vmflags(char *buf, struct vma_area *vma_area)
 #define _vmflag_match(_t, _s) (_t[0] == _s[0] && _t[1] == _s[1])
 
 	do {
+		/* open() block */
+		if (_vmflag_match(tok, "sh"))
+			shared = true;
+		else if (_vmflag_match(tok, "mw"))
+			maywrite = true;
+
 		/* mmap() block */
 		if (_vmflag_match(tok, "gd"))
 			vma_area->e->flags |= MAP_GROWSDOWN;
@@ -129,6 +141,18 @@ static int parse_vmflags(char *buf, struct vma_area *vma_area)
 		else if (_vmflag_match(tok, "nh"))
 			vma_area->e->madv |= (1ul << MADV_NOHUGEPAGE);
 
+		/* vmsplice doesn't work for VM_IO and VM_PFNMAP mappings. */
+		if (_vmflag_match(tok, "io") || _vmflag_match(tok, "pf")) {
+#ifdef CONFIG_VDSO
+			/*
+			 * VVAR area mapped by the kernel as
+			 * VM_IO | VM_PFNMAP| VM_DONTEXPAND | VM_DONTDUMP
+			 */
+			if (!vma_area_is(vma_area, VMA_AREA_VVAR))
+#endif
+				vma_area->e->status |= VMA_UNSUPP;
+		}
+
 		/*
 		 * Anything else is just ignored.
 		 */
@@ -136,6 +160,12 @@ static int parse_vmflags(char *buf, struct vma_area *vma_area)
 
 #undef _vmflag_match
 
+	if (shared && maywrite)
+		vma_area->e->fdflags = O_RDWR;
+	else
+		vma_area->e->fdflags = O_RDONLY;
+	vma_area->e->has_fdflags = true;
+
 	if (vma_area->e->madv)
 		vma_area->e->has_madv = true;
 
@@ -172,6 +202,13 @@ static int vma_get_mapfile(struct vma_area *vma, DIR *mfd,
 	if (prev_vfi->vma && vfi_equal(vfi, prev_vfi)) {
 		struct vma_area *prev = prev_vfi->vma;
 
+		/*
+		 * If vfi is equal (!) and negative @vm_file_fd --
+		 * we have nothing to borrow for sure.
+		 */
+		if (prev->vm_file_fd < 0)
+			return 0;
+
 		pr_debug("vma %"PRIx64" borrows vfi from previous %"PRIx64"\n",
 				vma->e->start, prev->e->start);
 		vma->vm_file_fd = prev->vm_file_fd;
@@ -207,7 +244,8 @@ static int vma_get_mapfile(struct vma_area *vma, DIR *mfd,
 			vma->vm_socket_id = buf.st_ino;
 		} else if (errno != ENOENT)
 			return -1;
-	}
+	} else if (opts.aufs && fixup_aufs_vma_fd(vma) < 0)
+		return -1;
 
 	return 0;
 }
@@ -218,7 +256,7 @@ int parse_self_maps_lite(struct vm_area_list *vms)
 
 	vm_area_list_init(vms);
 
-	maps = fopen("/proc/self/maps", "r");
+	maps = fopen_proc(PROC_SELF, "maps");
 	if (maps == NULL) {
 		pr_perror("Can't open self maps");
 		return -1;
@@ -251,8 +289,7 @@ static char smaps_buf[PAGE_SIZE];
 int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_files)
 {
 	struct vma_area *vma_area = NULL;
-	unsigned long start, end, pgoff;
-	bool prev_growsdown = false;
+	unsigned long start, end, pgoff, prev_end = 0;
 	char r, w, x, s;
 	int ret = -1;
 	struct vma_file_info vfi;
@@ -306,10 +343,17 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 		}
 
 		if (vma_area) {
-			/* If we've split the stack vma, only the lowest one has the guard page. */
-			if ((vma_area->e->flags & MAP_GROWSDOWN) && !prev_growsdown)
+			if (vma_area->e->status & VMA_UNSUPP) {
+				pr_err("Unsupported mapping found %016"PRIx64"-%016"PRIx64"\n",
+							vma_area->e->start, vma_area->e->end);
+				goto err;
+			}
+
+			/* Add a guard page only if here is enough space for it */
+			if ((vma_area->e->flags & MAP_GROWSDOWN) &&
+			    prev_end < vma_area->e->start)
 				vma_area->e->start -= PAGE_SIZE; /* Guard page */
-			prev_growsdown = (bool)(vma_area->e->flags & MAP_GROWSDOWN);
+			prev_end = vma_area->e->end;
 
 			list_add_tail(&vma_area->list, &vma_area_list->h);
 			vma_area_list->nr++;
@@ -370,9 +414,23 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 		} else if (strstr(buf, "[vsyscall]") || strstr(buf, "[vectors]")) {
 			vma_area->e->status |= VMA_AREA_VSYSCALL;
 		} else if (strstr(buf, "[vdso]")) {
+#ifdef CONFIG_VDSO
 			vma_area->e->status |= VMA_AREA_REGULAR;
 			if ((vma_area->e->prot & VDSO_PROT) == VDSO_PROT)
 				vma_area->e->status |= VMA_AREA_VDSO;
+#else
+			pr_warn_once("Found vDSO area without support\n");
+			goto err;
+#endif
+		} else if (strstr(buf, "[vvar]")) {
+#ifdef CONFIG_VDSO
+			vma_area->e->status |= VMA_AREA_REGULAR;
+			if ((vma_area->e->prot & VVAR_PROT) == VVAR_PROT)
+				vma_area->e->status |= VMA_AREA_VVAR;
+#else
+			pr_warn_once("Found VVAR area without support\n");
+			goto err;
+#endif
 		} else if (strstr(buf, "[heap]")) {
 			vma_area->e->status |= VMA_AREA_REGULAR | VMA_AREA_HEAP;
 		} else {
@@ -402,7 +460,19 @@ int parse_smaps(pid_t pid, struct vm_area_list *vma_area_list, bool use_map_file
 			if (!st_buf)
 				goto err;
 
-			if (fstat(vma_area->vm_file_fd, st_buf) < 0) {
+			/*
+			 * For AUFS support, we cannot fstat() a file descriptor that
+			 * is a symbolic link to a branch (it would return different
+			 * dev/ino than the real file).  Instead, we stat() using the
+			 * full pathname that we saved before.
+			 */
+			if (vma_area->aufs_fpath) {
+				if (stat(vma_area->aufs_fpath, st_buf) < 0) {
+					pr_perror("Failed stat on %d's map %lu (%s)",
+						pid, start, vma_area->aufs_fpath);
+					goto err;
+				}
+			} else if (fstat(vma_area->vm_file_fd, st_buf) < 0) {
 				pr_perror("Failed fstat on %d's map %lu", pid, start);
 				goto err;
 			}
@@ -836,13 +906,22 @@ static int parse_mountinfo_ent(char *str, struct mount_info *new)
 	char *opt;
 	char *fstype;
 
-	ret = sscanf(str, "%i %i %u:%u %ms %ms %ms %n",
-			&new->mnt_id, &new->parent_mnt_id,
-			&kmaj, &kmin, &new->root, &new->mountpoint,
-			&opt, &n);
-	if (ret != 7)
+	new->mountpoint = xmalloc(PATH_MAX);
+	if (new->mountpoint == NULL)
 		return -1;
 
+	new->mountpoint[0] = '.';
+	ret = sscanf(str, "%i %i %u:%u %ms %s %ms %n",
+			&new->mnt_id, &new->parent_mnt_id,
+			&kmaj, &kmin, &new->root, new->mountpoint + 1,
+			&opt, &n);
+	if (ret != 7) {
+		xfree(new->mountpoint);
+		return -1;
+	}
+
+	new->mountpoint = xrealloc(new->mountpoint, strlen(new->mountpoint) + 1);
+
 	new->s_dev = MKKDEV(kmaj, kmin);
 	new->flags = 0;
 	if (parse_mnt_flags(opt, &new->flags))
@@ -876,14 +955,13 @@ err:
 	return ret;
 }
 
-struct mount_info *parse_mountinfo(pid_t pid)
+struct mount_info *parse_mountinfo(pid_t pid, struct ns_id *nsid)
 {
 	struct mount_info *list = NULL;
 	FILE *f;
 	char str[1024];
 
-	snprintf(str, sizeof(str), "/proc/%d/mountinfo", pid);
-	f = fopen(str, "r");
+	f = fopen_proc(pid, "mountinfo");
 	if (!f) {
 		pr_perror("Can't open %d mountinfo", pid);
 		return NULL;
@@ -897,6 +975,8 @@ struct mount_info *parse_mountinfo(pid_t pid)
 		if (!new)
 			goto err;
 
+		new->nsid = nsid;
+
 		new->next = list;
 		list = new;
 
@@ -979,6 +1059,51 @@ static void parse_fhandle_encoded(char *tok, FhEntry *fh)
 	}
 }
 
+static int parse_timerfd(FILE *f, char *buf, size_t size, TimerfdEntry *tfy)
+{
+	/*
+	 * Format is
+	 * clockid: 0
+	 * ticks: 0
+	 * settime flags: 01
+	 * it_value: (0, 49406829)
+	 * it_interval: (1, 0)
+	 */
+	if (sscanf(buf, "clockid: %d", &tfy->clockid) != 1)
+		goto parse_err;
+
+	if (!fgets(buf, size, f))
+		goto nodata;
+	if (sscanf(buf, "ticks: %llu", (unsigned long long *)&tfy->ticks) != 1)
+		goto parse_err;
+
+	if (!fgets(buf, size, f))
+		goto nodata;
+	if (sscanf(buf, "settime flags: 0%o", &tfy->settime_flags) != 1)
+		goto parse_err;
+
+	if (!fgets(buf, size, f))
+		goto nodata;
+	if (sscanf(buf, "it_value: (%llu, %llu)",
+		   (unsigned long long *)&tfy->vsec,
+		   (unsigned long long *)&tfy->vnsec) != 2)
+		goto parse_err;
+
+	if (!fgets(buf, size, f))
+		goto nodata;
+	if (sscanf(buf, "it_interval: (%llu, %llu)",
+		   (unsigned long long *)&tfy->isec,
+		   (unsigned long long *)&tfy->insec) != 2)
+		goto parse_err;
+	return 0;
+
+parse_err:
+	return -1;
+nodata:
+	pr_err("No data left in proc file while parsing timerfd\n");
+	goto parse_err;
+}
+
 #define fdinfo_field(str, field)	!strncmp(str, field":", sizeof(field))
 
 static int parse_fdinfo_pid_s(char *pid, int fd, int type,
@@ -999,7 +1124,30 @@ static int parse_fdinfo_pid_s(char *pid, int fd, int type,
 	while (fgets(str, sizeof(str), f)) {
 		union fdinfo_entries entry;
 
-		if (fdinfo_field(str, "pos") || fdinfo_field(str, "counter"))
+		if (fdinfo_field(str, "pos") ||
+		    fdinfo_field(str, "flags") ||
+		    fdinfo_field(str, "mnt_id")) {
+			unsigned long long val;
+			struct fdinfo_common *fdinfo = arg;
+
+			if (type != FD_TYPES__UND)
+				continue;
+			ret = sscanf(str, "%*s %lli", &val);
+			if (ret != 1)
+				goto parse_err;
+
+			if (fdinfo_field(str, "pos"))
+				fdinfo->pos = val;
+			else if (fdinfo_field(str, "flags"))
+				fdinfo->flags = val;
+			else if (fdinfo_field(str, "mnt_id"))
+				fdinfo->mnt_id = val;
+
+			entry_met = true;
+			continue;
+		}
+
+		if (type == FD_TYPES__UND)
 			continue;
 
 		if (fdinfo_field(str, "eventfd-count")) {
@@ -1018,6 +1166,21 @@ static int parse_fdinfo_pid_s(char *pid, int fd, int type,
 			entry_met = true;
 			continue;
 		}
+		if (fdinfo_field(str, "clockid")) {
+			timerfd_entry__init(&entry.tfy);
+
+			if (type != FD_TYPES__TIMERFD)
+				goto parse_err;
+			ret = parse_timerfd(f, str, sizeof(str), &entry.tfy);
+			if (ret)
+				goto parse_err;
+			ret = cb(&entry, arg);
+			if (ret)
+				goto out;
+
+			entry_met = true;
+			continue;
+		}
 		if (fdinfo_field(str, "tfd")) {
 			eventpoll_tfd_entry__init(&entry.epl);
 
@@ -1066,7 +1229,7 @@ static int parse_fdinfo_pid_s(char *pid, int fd, int type,
 		if (fdinfo_field(str, "fanotify ino")) {
 			FanotifyInodeMarkEntry ie = FANOTIFY_INODE_MARK_ENTRY__INIT;
 			FhEntry f_handle = FH_ENTRY__INIT;
-			int hoff;
+			int hoff = 0;
 
 			if (type != FD_TYPES__FANOTIFY)
 				goto parse_err;
@@ -1082,7 +1245,7 @@ static int parse_fdinfo_pid_s(char *pid, int fd, int type,
 				     &entry.ffy.mflags, &entry.ffy.mask, &entry.ffy.ignored_mask,
 				     &f_handle.bytes, &f_handle.type,
 				     &hoff);
-			if (ret != 7)
+			if (ret != 7 || hoff == 0)
 				goto parse_err;
 
 			if (alloc_fhandle(&f_handle)) {
@@ -1293,6 +1456,16 @@ err:
 	return ret;
 }
 
+void free_posix_timers(struct proc_posix_timers_stat *st)
+{
+	while (!list_empty(&st->timers)) {
+		struct proc_posix_timer *timer;
+		timer = list_first_entry(&st->timers, struct proc_posix_timer, list);
+		list_del(&timer->list);
+		xfree(timer);
+	}
+}
+
 int parse_posix_timers(pid_t pid, struct proc_posix_timers_stat *args)
 {
 	int ret = 0;
@@ -1365,11 +1538,7 @@ int parse_posix_timers(pid_t pid, struct proc_posix_timers_stat *args)
 		args->timer_n++;
 	}
 err:
-	while (!list_empty(&args->timers)) {
-		timer = list_first_entry(&args->timers, struct proc_posix_timer, list);
-		list_del(&timer->list);
-		xfree(timer);
-	}
+	free_posix_timers(args);
 	pr_perror("Parse error in posix timers proc file!");
 	ret = -1;
 out:
@@ -1421,3 +1590,164 @@ int parse_threads(int pid, struct pid **_t, int *_n)
 
 	return 0;
 }
+
+int parse_task_cgroup(int pid, struct list_head *retl, unsigned int *n)
+{
+	int ret = 0;
+	FILE *f;
+
+	f = fopen_proc(pid, "cgroup");
+	while (fgets(buf, BUF_SIZE, f)) {
+		struct cg_ctl *ncc, *cc;
+		char *name, *path = NULL, *e;
+
+		ret = -1;
+		ncc = xmalloc(sizeof(*cc));
+		if (!ncc)
+			goto err;
+
+		/*
+		 * Typical output (':' is a separator here)
+		 *
+		 * 4:cpu,cpuacct:/
+		 * 3:cpuset:/
+		 * 2:name=systemd:/user.slice/user-1000.slice/session-1.scope
+		 */
+		name = strchr(buf, ':') + 1;
+		if (name)
+			path = strchr(name, ':');
+		if (!name || !path) {
+			pr_err("Failed parsing cgroup %s\n", buf);
+			xfree(ncc);
+			goto err;
+		}
+		e = strchr(name, '\n');
+		*path++ = '\0';
+		if (e)
+			*e = '\0';
+
+		ncc->name = xstrdup(name);
+		ncc->path = xstrdup(path);
+		if (!ncc->name || !ncc->path) {
+			xfree(ncc->name);
+			xfree(ncc->path);
+			xfree(ncc);
+			goto err;
+		}
+
+		list_for_each_entry(cc, retl, l)
+			if (strcmp(cc->name, name) >= 0 && strcmp(cc->path, path) >= 0)
+				break;
+
+		list_add_tail(&ncc->l, &cc->l);
+		(*n)++;
+	}
+
+	fclose(f);
+	return 0;
+
+err:
+	put_ctls(retl);
+	fclose(f);
+	return ret;
+}
+
+void put_ctls(struct list_head *l)
+{
+	struct cg_ctl *c, *n;
+
+	list_for_each_entry_safe(c, n, l, l) {
+		xfree(c->name);
+		xfree(c->path);
+		xfree(c);
+	}
+}
+
+
+/* Parse and create all the real controllers. This does not include things with
+ * the "name=" prefix, e.g. systemd.
+ */
+int parse_cgroups(struct list_head *cgroups, unsigned int *n_cgroups)
+{
+	FILE *f;
+	char buf[1024], name[1024];
+	int heirarchy, ret = 0;
+	struct cg_controller *cur = NULL;
+
+	f = fopen("/proc/cgroups", "r");
+	if (!f) {
+		pr_perror("failed opening /proc/cgroups");
+		return -1;
+	}
+
+	/* throw away the header */
+	if (!fgets(buf, 1024, f)) {
+		ret = -1;
+		goto out;
+	}
+
+	while (fgets(buf, 1024, f)) {
+		char *n;
+		char found = 0;
+
+		sscanf(buf, "%s %d", name, &heirarchy);
+		list_for_each_entry(cur, cgroups, l) {
+			if (cur->heirarchy == heirarchy) {
+				void *m;
+
+				found = 1;
+				cur->n_controllers++;
+				m = xrealloc(cur->controllers, sizeof(char *) * cur->n_controllers);
+				if (!m) {
+					ret = -1;
+					goto out;
+				}
+
+				cur->controllers = m;
+				if (!cur->controllers) {
+					ret = -1;
+					goto out;
+				}
+
+				n = xstrdup(name);
+				if (!n) {
+					ret = -1;
+					goto out;
+				}
+
+				cur->controllers[cur->n_controllers-1] = n;
+				break;
+			}
+		}
+
+		if (!found) {
+			struct cg_controller *nc = new_controller(name, heirarchy);
+			if (!nc) {
+				ret = -1;
+				goto out;
+			}
+			list_add_tail(&nc->l, &cur->l);
+			(*n_cgroups)++;
+		}
+	}
+
+out:
+	fclose(f);
+	return ret;
+}
+
+/*
+ * AUFS callback function to "fix up" the root pathname.
+ * See sysfs_parse.c for details.
+ */
+int aufs_parse(struct mount_info *new)
+{
+	int ret = 0;
+
+	if (!strcmp(new->mountpoint, "./")) {
+		opts.aufs = true;
+		ret = parse_aufs_branches(new);
+	}
+
+	return ret;
+}
